"""Top-level package for unasync."""

import collections
import errno
import os
import sys
import tokenize as std_tokenize

import tokenize_rt
from setuptools.command import build_py as orig

from ._version import __version__  # NOQA

__all__ = [
    "Rule",
    "unasync_files",
    "cmdclass_build_py",
]


_ASYNC_TO_SYNC = {
    "__aenter__": "__enter__",
    "__aexit__": "__exit__",
    "__aiter__": "__iter__",
    "__anext__": "__next__",
    "asynccontextmanager": "contextmanager",
    "AsyncIterable": "Iterable",
    "AsyncIterator": "Iterator",
    "AsyncGenerator": "Generator",
    # TODO StopIteration is still accepted in Python 2, but the right change
    # is 'raise StopAsyncIteration' -> 'return' since we want to use unasynced
    # code in Python 3.7+
    "StopAsyncIteration": "StopIteration",
}


class Rule:
    """A single set of rules for 'unasync'ing file(s)"""

    def __init__(self, fromdir, todir, additional_replacements=None):
        self.fromdir = fromdir.replace("/", os.sep)
        self.todir = todir.replace("/", os.sep)

        # Add any additional user-defined token replacements to our list.
        self.token_replacements = _ASYNC_TO_SYNC.copy()
        for key, val in (additional_replacements or {}).items():
            self.token_replacements[key] = val

    def _match(self, filepath):
        """Determines if a Rule matches a given filepath and if so
        returns a higher comparable value if the match is more specific.
        """
        file_segments = [x for x in filepath.split(os.sep) if x]
        from_segments = [x for x in self.fromdir.split(os.sep) if x]
        len_from_segments = len(from_segments)

        if len_from_segments > len(file_segments):
            return False

        for i in range(len(file_segments) - len_from_segments + 1):
            if file_segments[i : i + len_from_segments] == from_segments:
                return len_from_segments, i

        return False

    def _unasync_file(self, filepath):
        with open(filepath, "rb") as f:
            encoding, _ = std_tokenize.detect_encoding(f.readline)

        with open(filepath, encoding=encoding) as f:
            tokens = tokenize_rt.src_to_tokens(f.read())
            tokens = self._unasync_tokens(tokens)
            result = tokenize_rt.tokens_to_src(tokens)
            outfilepath = filepath.replace(self.fromdir, self.todir)
            os.makedirs(os.path.dirname(outfilepath), exist_ok=True)
            with open(outfilepath, "wb") as f:
                f.write(result.encode(encoding))

    def _unasync_tokens(self, tokens):
        skip_next = False
        for i, token in enumerate(tokens):
            if skip_next:
                skip_next = False
                continue

            if token.src in ["async", "await"]:
                # When removing async or await, we want to skip the following whitespace
                # so that `print(await stuff)` becomes `print(stuff)` and not `print( stuff)`
                skip_next = True
            else:
                if token.name == "NAME":
                    token = token._replace(src=self._unasync_name(token.src))
                elif token.name == "STRING":
                    left_quote, name, right_quote = (
                        token.src[0],
                        token.src[1:-1],
                        token.src[-1],
                    )
                    token = token._replace(
                        src=left_quote + self._unasync_name(name) + right_quote
                    )

                yield token

    def _unasync_name(self, name):
        if name in self.token_replacements:
            return self.token_replacements[name]
        # Convert classes prefixed with 'Async' into 'Sync'
        elif len(name) > 5 and name.startswith("Async") and name[5].isupper():
            return "Sync" + name[5:]
        return name


def unasync_files(fpath_list, rules):
    for f in fpath_list:
        found_rule = None
        found_weight = None

        for rule in rules:
            weight = rule._match(f)
            if weight and (found_weight is None or weight > found_weight):
                found_rule = rule
                found_weight = weight

        if found_rule:
            found_rule._unasync_file(f)


Token = collections.namedtuple("Token", ["type", "string", "start", "end", "line"])


_DEFAULT_RULE = Rule(fromdir="/_async/", todir="/_sync/")


class _build_py(orig.build_py):
    """
    Subclass build_py from setuptools to modify its behavior.

    Convert files in _async dir from being asynchronous to synchronous
    and saves them in _sync dir.
    """

    UNASYNC_RULES = (_DEFAULT_RULE,)

    def run(self):
        rules = self.UNASYNC_RULES

        self._updated_files = []

        # Base class code
        if self.py_modules:
            self.build_modules()
        if self.packages:
            self.build_packages()
            self.build_package_data()

        # Our modification!
        unasync_files(self._updated_files, rules)

        # Remaining base class code
        self.byte_compile(self.get_outputs(include_bytecode=0))

    def build_module(self, module, module_file, package):
        outfile, copied = super().build_module(module, module_file, package)
        if copied:
            self._updated_files.append(outfile)
        return outfile, copied


def cmdclass_build_py(rules=(_DEFAULT_RULE,)):
    """Creates a 'build_py' class for use within 'cmdclass={"build_py": ...}'"""

    class _custom_build_py(_build_py):
        UNASYNC_RULES = rules

    return _custom_build_py



from contextlib import asynccontextmanager


class TestImplementation:
    async def __aenter__(self):
        return self

    async def __aexit__(self):
        await self.close()

    @asynccontextmanager
    async def context_manager(self):
        return self



class TestImplementation:
    def __init__(self, a, b):
        self.a = a
        self.b = b
        self.s = "UTF-8: ❄"

    def get_a_b(self):
        # fmt: off
        s = "a is %s b is %s" % \
            (self.a,
             self.b)
        # fmt: on
        return s

    def f(self):
        return 1


def f():
    return 1



from contextlib import contextmanager


class TestImplementation:
    def __enter__(self):
        return self

    def __exit__(self):
        self.close()

    @contextmanager
    def context_manager(self):
        return self



import setuptools

import unasync

setuptools.setup(
    name="ahip",
    version="0.0.1",
    author="Example Author",
    author_email="author@example.com",
    description="A package used to test customized unasync",
    url="https://github.com/pypa/sampleproject",
    packages=["ahip", "ahip.some_dir", "ahip.tests"],
    cmdclass={
        "build_py": unasync.cmdclass_build_py(
            rules=[
                unasync.Rule(fromdir="/ahip/", todir="/hip/"),
                unasync.Rule(
                    fromdir="/ahip/tests/",
                    todir="/hip/tests/",
                    additional_replacements={"ahip": "hip"},
                ),
            ]
        )
    },
    package_dir={"": "src"},
)



import setuptools

import unasync

setuptools.setup(
    name="example_pkg",
    version="0.0.1",
    author="Example Author",
    author_email="author@example.com",
    description="A package used to test unasync",
    url="https://github.com/pypa/sampleproject",
    packages=["example_pkg", "example_pkg._async", "example_pkg._async.some_dir"],
    cmdclass={"build_py": unasync.cmdclass_build_py()},
    package_dir={"": "src"},
)



import copy
import errno
import os
import shutil
import subprocess

import pytest

import unasync

TEST_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
ASYNC_DIR = os.path.join(TEST_DIR, "async")
SYNC_DIR = os.path.join(TEST_DIR, "sync")
TEST_FILES = sorted(f for f in os.listdir(ASYNC_DIR) if f.endswith(".py"))


def list_files(startpath):
    output = ""
    for root, dirs, files in os.walk(startpath):
        level = root.replace(startpath, "").count(os.sep)
        indent = " " * 4 * (level)
        output += f"{indent}{os.path.basename(root)}/"
        output += "\n"
        subindent = " " * 4 * (level + 1)
        for f in files:
            output += f"{subindent}{f}"
            output += "\n"
    return output


def test_rule_on_short_path():
    rule = unasync.Rule("/ahip/tests/", "/hip/tests/")
    assert rule._match("/ahip/") is False


@pytest.mark.parametrize("source_file", TEST_FILES)
def test_unasync(tmpdir, source_file):

    rule = unasync.Rule(fromdir=ASYNC_DIR, todir=str(tmpdir))
    rule._unasync_file(os.path.join(ASYNC_DIR, source_file))

    encoding = "latin-1" if "encoding" in source_file else "utf-8"
    with open(os.path.join(SYNC_DIR, source_file), encoding=encoding) as f:
        truth = f.read()
    with open(os.path.join(str(tmpdir), source_file), encoding=encoding) as f:
        unasynced_code = f.read()
        assert unasynced_code == truth


def test_unasync_files(tmpdir):
    """Test the unasync_files API, not tied by a Rule or to setuptools."""
    unasync.unasync_files(
        [os.path.join(ASYNC_DIR, fpath) for fpath in TEST_FILES],
        rules=[unasync.Rule(fromdir=ASYNC_DIR, todir=str(tmpdir))],
    )

    for source_file in TEST_FILES:
        encoding = "latin-1" if "encoding" in source_file else "utf-8"
        with open(os.path.join(SYNC_DIR, source_file), encoding=encoding) as f:
            truth = f.read()
        with open(os.path.join(str(tmpdir), source_file), encoding=encoding) as f:
            unasynced_code = f.read()
            assert unasynced_code == truth


def test_build_py_modules(tmpdir):

    source_modules_dir = os.path.join(TEST_DIR, "example_mod")
    mod_dir = str(tmpdir) + "/" + "example_mod"
    shutil.copytree(source_modules_dir, mod_dir)

    env = copy.copy(os.environ)
    env["PYTHONPATH"] = os.path.realpath(os.path.join(TEST_DIR, ".."))
    subprocess.check_call(["python", "setup.py", "build"], cwd=mod_dir, env=env)
    # Calling it twice to test the "if not copied" branch
    subprocess.check_call(["python", "setup.py", "build"], cwd=mod_dir, env=env)

    unasynced = os.path.join(mod_dir, "build/lib/_sync/some_file.py")
    tree_build_dir = list_files(mod_dir)

    with open(unasynced) as f:
        unasynced_code = f.read()
        assert unasynced_code == "def f():\n    return 1\n"


def test_build_py_packages(tmpdir):

    source_pkg_dir = os.path.join(TEST_DIR, "example_pkg")
    pkg_dir = str(tmpdir) + "/" + "example_pkg"
    shutil.copytree(source_pkg_dir, pkg_dir)

    env = copy.copy(os.environ)
    env["PYTHONPATH"] = os.path.realpath(os.path.join(TEST_DIR, ".."))
    subprocess.check_call(["python", "setup.py", "build"], cwd=pkg_dir, env=env)

    unasynced = os.path.join(pkg_dir, "build/lib/example_pkg/_sync/__init__.py")

    with open(unasynced) as f:
        unasynced_code = f.read()
        assert unasynced_code == "def f():\n    return 1\n"


def test_project_structure_after_build_py_packages(tmpdir):

    source_pkg_dir = os.path.join(TEST_DIR, "example_pkg")
    pkg_dir = str(tmpdir) + "/" + "example_pkg"
    shutil.copytree(source_pkg_dir, pkg_dir)

    env = copy.copy(os.environ)
    env["PYTHONPATH"] = os.path.realpath(os.path.join(TEST_DIR, ".."))
    subprocess.check_call(["python", "setup.py", "build"], cwd=pkg_dir, env=env)

    _async_dir_tree = list_files(
        os.path.join(source_pkg_dir, "src/example_pkg/_async/.")
    )
    unasynced_dir_tree = list_files(
        os.path.join(pkg_dir, "build/lib/example_pkg/_sync/.")
    )

    assert _async_dir_tree == unasynced_dir_tree


def test_project_structure_after_customized_build_py_packages(tmpdir):

    source_pkg_dir = os.path.join(TEST_DIR, "example_custom_pkg")
    pkg_dir = str(tmpdir) + "/" + "example_custom_pkg"
    shutil.copytree(source_pkg_dir, pkg_dir)

    env = copy.copy(os.environ)
    env["PYTHONPATH"] = os.path.realpath(os.path.join(TEST_DIR, ".."))
    subprocess.check_call(["python", "setup.py", "build"], cwd=pkg_dir, env=env)

    _async_dir_tree = list_files(os.path.join(source_pkg_dir, "src/ahip/."))
    unasynced_dir_path = os.path.join(pkg_dir, "build/lib/hip/.")
    unasynced_dir_tree = list_files(unasynced_dir_path)

    assert _async_dir_tree == unasynced_dir_tree

    with open(os.path.join(unasynced_dir_path, "tests/test_conn.py")) as f:
        assert "import hip\n" in f.read()



from setuptools import find_packages, setup

exec(open("src/unasync/_version.py", encoding="utf-8").read())

LONG_DESC = open("README.rst", encoding="utf-8").read()

setup(
    name="unasync",
    version=__version__,
    description="The async transformation code.",
    url="https://github.com/python-trio/unasync",
    long_description=LONG_DESC,
    long_description_content_type="text/x-rst",
    author="Ratan Kulshreshtha",
    author_email="ratan.shreshtha@gmail.com",
    license="MIT OR Apache-2.0",
    include_package_data=True,
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=["tokenize_rt", "setuptools"],
    keywords=["async"],
    python_requires=">=3.8",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Framework :: Trio",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



class AsyncLock:
    ...


class AsyncSocket:
    def __init__(self, send_lock: AsyncLock):
        ...

    async def __aenter__(self) -> "AsyncSocket":
        ...

    async def send_all(self, data):
        ...
